/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 26/06/22.
//

#include <cstdlib>

#define FGGL_PLATFORM_PATHS linux

#include "fggl/platform/linux/paths.hpp"
#include "fggl/debug/logging.hpp"

namespace fggl::platform {

	inline static std::filesystem::path get_user_path(const char *env, const char *fallback) {
		const char *path = std::getenv(env);
		if (path != nullptr) {
			return {path};
		}
		return {fallback};
	}

	static std::vector<std::filesystem::path> get_path_list(const char *env, const char *folderName) {
		const char *pathList = std::getenv(env);
		std::vector<std::filesystem::path> paths;
		if (pathList) {
			std::string pathListStr(pathList);
			std::string::size_type pos = 0;
			while (pos < pathListStr.size()) {
				std::string::size_type nextPos = pathListStr.find(':', pos);
				if (nextPos == std::string::npos) {
					nextPos = pathListStr.size();
				}
				std::string path = pathListStr.substr(pos, nextPos - pos);
				paths.push_back(std::filesystem::path(path) / folderName);
				pos = nextPos + 1;
			}
		}
		return paths;
	}

	EnginePaths calc_engine_paths(const char *base) {
		auto dataDirs = get_path_list(ENV_DATA_DIRS, base);
		if (dataDirs.empty()) {
			for (const auto &defaultDir : DEFAULT_DATA_DIRS) {
				dataDirs.push_back(std::filesystem::path(defaultDir) / base);
			}
		}

		auto configDirs = get_path_list(ENV_CONFIG_DIRS, base);
		if (configDirs.empty()) {
			for (const auto &defaultDir : DEFAULT_CONFIG_DIRS) {
				configDirs.push_back(std::filesystem::path(defaultDir) / base);
			}
		}

		return EnginePaths{
			get_user_path(ENV_USER_CONFIG, DEFAULT_USER_CONFIG) / base,
			get_user_path(ENV_USER_DATA, DEFAULT_USER_DATA) / base,
			get_user_path(ENV_USER_CACHE, DEFAULT_USER_CACHE) / base,
			dataDirs,
			configDirs
		};
	}

	std::filesystem::path locate_data(const EnginePaths &paths, const std::filesystem::path &relPath) {
		auto userPath = paths.userData / relPath;
		if (std::filesystem::exists(userPath)) {
			return userPath;
		}

		// check system paths
		for (const auto &path : paths.dataDirs) {
			auto fullPath = path / relPath;
			debug::trace("Checking data path: {}, exists: {}", fullPath.c_str(), std::filesystem::exists(fullPath));
			if (std::filesystem::exists(fullPath)) {
				return fullPath;
			}
		}

		// if debug mode, try CWD as well.
		auto debugPath = std::filesystem::current_path() / "data" / relPath;
		if (std::filesystem::exists(debugPath)) {
			debug::trace("Checking debug path: {}, exists: {}", debugPath.c_str(), std::filesystem::exists(debugPath));
			return debugPath;
		}

		// if the file existed, it shoudl exist in the user space
		return userPath;
	}

	std::filesystem::path locate_config(const EnginePaths &paths, const std::filesystem::path &relPath) {
		auto userPath = paths.userConfig / relPath;
		if (std::filesystem::exists(userPath)) {
			return userPath;
		}

		// check system paths
		for (const auto &path : paths.configDirs) {
			auto fullPath = path / relPath;
			if (std::filesystem::exists(fullPath)) {
				return fullPath;
			}
		}

		// if the file existed, it should exist in the user space
		return userPath;
	}

	std::filesystem::path locate_cache(const EnginePaths &paths, const std::filesystem::path &relPath) {
		auto userPath = paths.userCache / relPath;
		if (std::filesystem::exists(userPath)) {
			return userPath;
		}

		// check system paths
		for (const auto &path : paths.configDirs) {
			auto fullPath = path / relPath;
			if (std::filesystem::exists(fullPath)) {
				return fullPath;
			}
		}

		// if the file existed, it should exist in the user space
		return userPath;
	}
} // namespace fggl::platform::linux
