/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_DATA_MODEL_HPP
#define FGGL_DATA_MODEL_HPP

#include <tuple>
#include <vector>
#include <string>
#include <climits>

#include <fggl/math/types.hpp>

namespace fggl::data {

	constexpr math::vec3 ILLEGAL_NORMAL{0.0F, 0.0F, 0.F};
	constexpr math::vec3 DEFAULT_COLOUR{1.0F, 1.0F, 1.0F};

	struct Vertex {
		math::vec3 posititon;
		math::vec3 normal;
		math::vec3 colour;
		math::vec2 texPos;

		inline static Vertex from_pos(math::vec3 pos) {
			return {
				pos,
				ILLEGAL_NORMAL,
				DEFAULT_COLOUR,
				{0.0F, 0.0F}
			};
		}
	};

	struct Vertex2D {
		fggl::math::vec2 position;
		fggl::math::vec3 colour;
		fggl::math::vec2 texPos;
	};

	struct Mesh2D {
		std::vector<Vertex2D> vertexList;
		std::vector<uint32_t> indexList;

		inline std::size_t add_vertex(const Vertex2D &vertex) {
			vertexList.push_back(vertex);
			return vertexList.size() - 1;
		}

		inline void add_index(uint32_t idx) {
			indexList.push_back(idx);
		}
	};


	// comparison operators

	inline bool operator<(const Vertex &lhs, const Vertex &rhs) {
		return std::tie(lhs.posititon, lhs.normal, lhs.colour)
			< std::tie(rhs.posititon, rhs.normal, rhs.colour);
	}

	inline bool operator==(const Vertex &lhs, const Vertex &rhs) {
		return lhs.posititon == rhs.posititon
			&& lhs.colour == rhs.colour
			&& lhs.normal == rhs.normal;
	}

	inline bool operator!=(const Vertex &lhs, const Vertex &rhs) {
		return !(lhs == rhs);
	}

	class Mesh {
		public:
			using IndexType = unsigned int;
			constexpr static const IndexType INVALID_IDX = IndexType(UINT_MAX);

			Mesh();
			~Mesh() = default;

			/**
			 * Push a vertex with deduplication.
			 *
			 * If the vertex is already known, the old vertex will be used and a new entry will
			 * be added to the indicies list.
			 */
			inline void push(Vertex vert) {
				auto idx = indexOf(vert);
				if (idx == INVALID_IDX) {
					idx = pushVertex(vert);
					pushIndex(idx);
				} else {
					pushIndex(idx);
				}
			}

			/**
			 * Push a new vertex index.
			 *
			 * Don't push non-existant indexes, it'll make the rendering backend sad.
			 */
			void pushIndex(IndexType idx);

			/**
			 * Force push a vertex duplicate.
			 *
			 * If the vetex already exists, it will be duplicated.
			 */
			IndexType pushVertex(Vertex vert);

			/**
			 * remove Duplicate verticies
			 *
			 * If generated using the procedural system, the mesh data can get a little messy.
			 * This method will remove unused and duplicate verticies.
			 */
			void removeDups();

			/**
			 * Search for a vertex.
			 *
			 * Returns the index of a vert, else returns -1.
			 */
			IndexType indexOf(Vertex vert);

			inline std::vector<Vertex> &vertexList() {
				return m_verts;
			}

			inline std::size_t vertexCount() {
				return m_verts.size();
			}

			inline std::vector<IndexType> &indexList() {
				return m_index;
			}

			inline std::size_t indexCount() {
				return m_index.size();
			}

			inline Vertex &vertex(IndexType idx) {
				return m_verts[idx];
			}

			IndexType restartVertex;
		private:
			std::vector<Vertex> m_verts;
			std::vector<IndexType> m_index;
	};

	struct StaticMesh {
		constexpr static const char name[] = "StaticMesh";
		constexpr static const util::GUID guid = util::make_guid("StaticMesh");

		data::Mesh mesh;
		std::string pipeline;

		inline StaticMesh() : mesh(), pipeline() {}

		inline StaticMesh(const data::Mesh &aMesh, std::string aPipeline) :
			mesh(aMesh), pipeline(std::move(aPipeline)) {}
	};

	class Model {
		public:
			Model() = default;
			~Model() = default;

			inline void append(const Mesh &mesh) {
				m_meshes.push_back(mesh);
			}

		private:
			std::vector<Mesh> m_meshes;
	};

}

#endif
